@@ -34,6 +34,7 @@
 #include <wlr/util/box.h>
 
 #include "cursor.h"
+#include "desktop-xwayland.h"
 #include "device-state.h"
 #include "idle-inhibit.h"
 #include "layer-shell.h"
@@ -55,7 +56,7 @@
 
 /* Maximum protocol versions we support */
 #define PHOC_FRACTIONAL_SCALE_VERSION 1
-#define PHOC_XDG_SHELL_VERSION 5
+#define PHOC_XDG_SHELL_VERSION 6
 #define PHOC_LAYER_SHELL_VERSION 3
 
 #define PHOC_ANIM_ALWAYS_ON_TOP_DURATION  300
@@ -188,7 +189,7 @@ desktop_view_at (PhocDesktop         *self,
   for (GList *l = phoc_desktop_get_views (self)->head; l; l = l->next) {
     PhocView *view = PHOC_VIEW (l->data);
 
-    if (phoc_desktop_view_is_visible (self, view) && view_at (view, lx, ly, surface, sx, sy))
+    if (phoc_desktop_view_check_visibility (self, view) && view_at (view, lx, ly, surface, sx, sy))
       return view;
   }
   return NULL;
@@ -303,11 +304,25 @@ phoc_desktop_wlr_surface_at (PhocDesktop *desktop,
   return NULL;
 }
 
+/**
+ * phoc_desktop_view_check_visibility:
+ * @self: The desktop
+ * @view: The view to check
+ *
+ * Checks if a view is currently visible. This is currently very
+ * pessimistic and only assumes that the view is not visible when
+ * we're certain it is covered by other windows.
+ *
+ * Returns: `FALSE` when it's certain that the view is not visible, otherwise `TRUE`
+ */
 gboolean
-phoc_desktop_view_is_visible (PhocDesktop *self, PhocView *view)
+phoc_desktop_view_check_visibility (PhocDesktop *self, PhocView *view)
 {
   PhocDesktopPrivate *priv;
+  PhocOutput *output;
   PhocView *top_view;
+  GQueue *layer_surfaces;
+  gboolean visible = TRUE;
 
   g_assert (PHOC_IS_DESKTOP (self));
   g_assert (PHOC_IS_VIEW (view));
@@ -315,42 +330,50 @@ phoc_desktop_view_is_visible (PhocDesktop *self, PhocView *view)
   priv = phoc_desktop_get_instance_private (self);
 
   if (!phoc_view_is_mapped (view)) {
-    return false;
+    visible = FALSE;
+    goto out;
   }
 
   g_assert_true (priv->views->head);
 
-  if (wl_list_length (&self->outputs) != 1) {
-    // current heuristics work well only for single output
-    return true;
+  /* current heuristics work well only for single output */
+  if (wl_list_length (&self->outputs) != 1)
+    goto out;
+
+  output = wl_container_of (self->outputs.next, output, link);
+  layer_surfaces = phoc_output_get_layer_surfaces_for_layer (output, ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY);
+  for (GList *l = layer_surfaces->head; l; l = l->next) {
+    PhocLayerSurface *layer_surface = PHOC_LAYER_SURFACE (l->data);
+
+    if (phoc_layer_surface_covers_output (layer_surface)) {
+      visible = FALSE;
+      goto out;
+    }
   }
 
-  if (!self->maximize) {
-    return true;
-  }
+  if (!self->maximize)
+    goto out;
 
   top_view = phoc_desktop_get_view_by_index (self, 0);
-#ifdef PHOC_XWAYLAND
-  // XWayland parent relations can be complicated and aren't described by PhocView
-  // relationships very well at the moment, so just make all XWayland windows visible
-  // when some XWayland window is active for now
-  if (PHOC_IS_XWAYLAND_SURFACE (view) && PHOC_IS_XWAYLAND_SURFACE (top_view)) {
-    return true;
-  }
-#endif
+  /* XWayland parent relations can be complicated and aren't described by PhocView
+   * relationships very well at the moment, so just make all XWayland windows visible
+   * when some XWayland window is active for now */
+  if (PHOC_IS_XWAYLAND_SURFACE (view) && PHOC_IS_XWAYLAND_SURFACE (top_view))
+    goto out;
+
+  for (PhocView *v = top_view; v; v = v->parent) {
+    if (v == view)
+      goto out;
 
-  PhocView *v = top_view;
-  while (v) {
-    if (v == view) {
-      return true;
-    }
     if (phoc_view_is_maximized (v)) {
-      return false;
+      visible = FALSE;
+      goto out;
     }
-    v = v->parent;
   }
 
-  return false;
+ out:
+  phoc_view_set_visibility (view, visible);
+  return visible;
 }
 
 static void
@@ -477,89 +500,6 @@ on_enable_animations_changed (PhocDesktop *self,
 }
 
 
-
-#ifdef PHOC_XWAYLAND
-static const char *atom_map[XWAYLAND_ATOM_LAST] = {
-        "_NET_WM_WINDOW_TYPE_NORMAL",
-        "_NET_WM_WINDOW_TYPE_DIALOG"
-};
-
-static void
-handle_xwayland_ready (struct wl_listener *listener,
-                       void               *data)
-{
-  PhocInput *input = phoc_server_get_input (phoc_server_get_default ());
-  PhocDesktop *desktop = wl_container_of (listener, desktop, xwayland_ready);
-  xcb_connection_t *xcb_conn = xcb_connect (NULL, NULL);
-
-  int err = xcb_connection_has_error (xcb_conn);
-  if (err) {
-    g_warning ("XCB connect failed: %d", err);
-    return;
-  }
-
-  xcb_intern_atom_cookie_t cookies[XWAYLAND_ATOM_LAST];
-
-  for (size_t i = 0; i < XWAYLAND_ATOM_LAST; i++)
-    cookies[i] = xcb_intern_atom (xcb_conn, 0, strlen (atom_map[i]), atom_map[i]);
-
-  for (size_t i = 0; i < XWAYLAND_ATOM_LAST; i++) {
-    xcb_generic_error_t *error = NULL;
-    xcb_intern_atom_reply_t *reply = xcb_intern_atom_reply (xcb_conn, cookies[i], &error);
-
-    if (error) {
-      g_warning ("could not resolve atom %s, X11 error code %d",
-                 atom_map[i], error->error_code);
-      free (error);
-    }
-
-    if (reply)
-      desktop->xwayland_atoms[i] = reply->atom;
-
-    free (reply);
-  }
-
-  xcb_disconnect (xcb_conn);
-
-#ifdef PHOC_XWAYLAND
-  if (desktop->xwayland != NULL) {
-    PhocSeat *xwayland_seat = phoc_input_get_seat (input, PHOC_CONFIG_DEFAULT_SEAT_NAME);
-    wlr_xwayland_set_seat (desktop->xwayland, xwayland_seat->seat);
-  }
-#endif
-
-}
-
-
-static void
-handle_xwayland_remove_startup_id (struct wl_listener *listener, void *data)
-{
-  PhocDesktop *desktop = wl_container_of (listener, desktop, xwayland_remove_startup_id);
-  struct wlr_xwayland_remove_startup_info_event *ev = data;
-
-  g_assert (PHOC_IS_DESKTOP (desktop));
-  g_assert (ev->id);
-
-  phoc_phosh_private_notify_startup_id (phoc_desktop_get_phosh_private (desktop),
-                                        ev->id,
-                                        PHOSH_PRIVATE_STARTUP_TRACKER_PROTOCOL_X11);
-}
-
-
-static void
-handle_xwayland_surface (struct wl_listener *listener, void *data)
-{
-  struct wlr_xwayland_surface *surface = data;
-  g_debug ("new xwayland surface: title=%s, class=%s, instance=%s",
-           surface->title, surface->class, surface->instance);
-  wlr_xwayland_surface_ping(surface);
-
-  /* Ref is dropped on surface destroy */
-  phoc_xwayland_surface_new (surface);
-}
-
-#endif /* PHOC_XWAYLAND */
-
 static void
 on_output_destroyed (PhocDesktop *self, PhocOutput *destroyed_output)
 {
@@ -597,59 +537,20 @@ handle_new_output (struct wl_listener *listener, void *data)
     return;
   }
 
-  g_signal_connect_swapped (output, "output-destroyed",
-                            G_CALLBACK (on_output_destroyed),
-                            self);
+  g_signal_connect_object (output, "output-destroyed",
+                           G_CALLBACK (on_output_destroyed),
+                           self,
+                           G_CONNECT_SWAPPED);
 }
 
 
 static void
-phoc_desktop_setup_xwayland (PhocDesktop *self)
+handle_backend_destroy (struct wl_listener *listener, void *data)
 {
-#ifdef PHOC_XWAYLAND
-  const char *cursor_default = PHOC_XCURSOR_DEFAULT;
-  PhocServer *server = phoc_server_get_default ();
-  PhocConfig *config = phoc_server_get_config (server);
+  PhocDesktop *self = wl_container_of (listener, self, backend_destroy);
 
-  self->xcursor_manager = wlr_xcursor_manager_create (NULL, PHOC_XCURSOR_SIZE);
-  g_return_if_fail (self->xcursor_manager);
-
-  if (config->xwayland) {
-    struct wl_display *wl_display = phoc_server_get_wl_display (server);
-    struct wlr_compositor *wlr_compositor = phoc_server_get_compositor (server);
-
-    self->xwayland = wlr_xwayland_create (wl_display, wlr_compositor, config->xwayland_lazy);
-    if (!self->xwayland) {
-      g_critical ("Failed to initialize Xwayland");
-      g_unsetenv ("DISPLAY");
-      return;
-    }
-
-    wl_signal_add (&self->xwayland->events.new_surface, &self->xwayland_surface);
-    self->xwayland_surface.notify = handle_xwayland_surface;
-
-    wl_signal_add (&self->xwayland->events.ready, &self->xwayland_ready);
-    self->xwayland_ready.notify = handle_xwayland_ready;
-
-    wl_signal_add (&self->xwayland->events.remove_startup_info, &self->xwayland_remove_startup_id);
-    self->xwayland_remove_startup_id.notify = handle_xwayland_remove_startup_id;
-
-    g_setenv ("DISPLAY", self->xwayland->display_name, true);
-
-    if (!wlr_xcursor_manager_load (self->xcursor_manager, 1))
-      g_critical ("Cannot load XWayland XCursor theme");
-
-    struct wlr_xcursor *xcursor = wlr_xcursor_manager_get_xcursor (self->xcursor_manager,
-                                                                   cursor_default,
-                                                                   1);
-    if (xcursor != NULL) {
-      struct wlr_xcursor_image *image = xcursor->images[0];
-      wlr_xwayland_set_cursor (self->xwayland, image->buffer,
-                               image->width * 4, image->width, image->height, image->hotspot_x,
-                               image->hotspot_y);
-    }
-  }
-#endif
+  wl_list_remove (&self->new_output.link);
+  wl_list_remove (&self->backend_destroy.link);
 }
 
 
@@ -669,14 +570,17 @@ phoc_desktop_constructed (GObject *object)
   self->new_output.notify = handle_new_output;
   wl_signal_add (&wlr_backend->events.new_output, &self->new_output);
 
-  self->layout = wlr_output_layout_create ();
+  self->backend_destroy.notify = handle_backend_destroy;
+  wl_signal_add (&wlr_backend->events.destroy, &self->backend_destroy);
+
+  self->layout = wlr_output_layout_create (wl_display);
   wlr_xdg_output_manager_v1_create (wl_display, self->layout);
   self->layout_change.notify = handle_layout_change;
   wl_signal_add (&self->layout->events.change, &self->layout_change);
 
   self->xdg_shell = wlr_xdg_shell_create(wl_display, PHOC_XDG_SHELL_VERSION);
-  wl_signal_add(&self->xdg_shell->events.new_surface, &self->xdg_shell_surface);
-  self->xdg_shell_surface.notify = phoc_handle_xdg_shell_surface;
+  wl_signal_add (&self->xdg_shell->events.new_toplevel, &self->xdg_shell_toplevel);
+  self->xdg_shell_toplevel.notify = phoc_handle_xdg_shell_toplevel;
 
   self->layer_shell = wlr_layer_shell_v1_create (wl_display, PHOC_LAYER_SHELL_VERSION);
   wl_signal_add(&self->layer_shell->events.new_surface, &self->layer_shell_surface);
@@ -746,7 +650,7 @@ phoc_desktop_constructed (GObject *object)
   self->pointer_constraint.notify = handle_pointer_constraint;
   wl_signal_add (&self->pointer_constraints->events.new_constraint, &self->pointer_constraint);
 
-  self->presentation = wlr_presentation_create (wl_display, wlr_backend);
+  wlr_presentation_create (wl_display, wlr_backend);
   self->foreign_toplevel_manager_v1 = wlr_foreign_toplevel_manager_v1_create (wl_display);
   self->relative_pointer_manager = wlr_relative_pointer_manager_v1_create (wl_display);
   self->pointer_gestures = wlr_pointer_gestures_v1_create (wl_display);
@@ -791,10 +695,9 @@ phoc_desktop_finalize (GObject *object)
 
   g_clear_pointer (&priv->views, g_queue_free);
 
-  /* TODO: currently destroys the backend before the desktop */
-  //wl_list_remove (&self->new_output.link);
+  wl_list_remove (&priv->gamma_control_set_gamma.link);
   wl_list_remove (&self->layout_change.link);
-  wl_list_remove (&self->xdg_shell_surface.link);
+  wl_list_remove (&self->xdg_shell_toplevel.link);
   wl_list_remove (&self->layer_shell_surface.link);
   wl_list_remove (&self->xdg_toplevel_decoration.link);
   wl_list_remove (&self->virtual_keyboard_new.link);
@@ -805,19 +708,7 @@ phoc_desktop_finalize (GObject *object)
   wl_list_remove (&self->output_power_manager_set_mode.link);
   wl_list_remove (&self->xdg_activation_v1_request_activate.link);
 
-#ifdef PHOC_XWAYLAND
-  /* Disconnect XWayland listener before shutting it down */
-  if (self->xwayland) {
-    wl_list_remove (&self->xwayland_surface.link);
-    wl_list_remove (&self->xwayland_ready.link);
-    wl_list_remove (&self->xwayland_remove_startup_id.link);
-  }
-
-  g_clear_pointer (&self->xcursor_manager, wlr_xcursor_manager_destroy);
-  // We need to shutdown Xwayland before disconnecting all clients, otherwise
-  // wlroots will restart it automatically.
-  g_clear_pointer (&self->xwayland, wlr_xwayland_destroy);
-#endif
+  phoc_desktop_destroy_xwayland (self);
 
   g_clear_pointer (&priv->idle_inhibit, phoc_idle_inhibit_destroy);
   g_clear_object (&priv->phosh);
